"""
Tests for `nameko_cachetools` module.
"""
import time
import pytest
from mock import Mock, patch
import random

import eventlet
from nameko.rpc import rpc
from nameko.standalone.rpc import ServiceRpcProxy
from nameko_cachetools import CachedRpcProxy, CacheFirstRpcProxy
from nameko.testing.services import (entrypoint_hook, entrypoint_waiter,
                                     get_extension)


@pytest.fixture
def container(container_factory, rabbit_config):
    class Service(object):
        name = "service"

        cached_service = CachedRpcProxy('some_other_service', failover_timeout=1)
        cache_first_service = CacheFirstRpcProxy('some_other_service')

        @rpc
        def cached(self, *args, **kwargs):
            return self.cached_service.some_method(*args, **kwargs)

        @rpc
        def cache_first(self, *args, **kwargs):
            return self.cache_first_service.some_method(*args, **kwargs)

    container = container_factory(Service, rabbit_config)
    container.start()

    return container


def test_cached_response(container):

    cached_rpc = get_extension(container, CachedRpcProxy)

    def fake_some_method(*args, **kwargs):
        return 'hi'

    with patch('nameko.rpc.MethodProxy.__call__', fake_some_method):
        with entrypoint_hook(container, 'cached') as hook:
            assert hook('test') == 'hi'

    def broken_some_method(*args, **kwargs):
        raise Exception('hmm')

    with patch('nameko.rpc.MethodProxy.__call__', broken_some_method):
        with entrypoint_hook(container, 'cached') as hook:
            assert hook('test') == 'hi'

    with patch('nameko.rpc.MethodProxy.__call__', broken_some_method):
        with entrypoint_hook(container, 'cached') as hook:
            with pytest.raises(Exception):
                hook('unknown')

    cached_rpc.cache = {}

    with patch('nameko.rpc.MethodProxy.__call__', broken_some_method):
        with entrypoint_hook(container, 'cached') as hook:
            with pytest.raises(Exception):
                hook('test')


def test_cached_response_on_timeout(container):

    cached_rpc = get_extension(container, CachedRpcProxy)

    def fake_some_method(*args, **kwargs):
        return 'hi'

    with patch('nameko.rpc.MethodProxy.__call__', fake_some_method):
        with entrypoint_hook(container, 'cached') as hook:
            assert hook() == 'hi'

    def slow_response(*args, **kwargs):
        eventlet.sleep(3)
        return 'hi'

    start = time.time()
    with patch('nameko.rpc.MethodProxy.__call__', slow_response):
        with entrypoint_hook(container, 'cached') as hook:
            assert hook() == 'hi'
    assert time.time() - start < 2

    cached_rpc.cache = {}

    start = time.time()
    with patch('nameko.rpc.MethodProxy.__call__', slow_response):
        with entrypoint_hook(container, 'cached') as hook:
            assert hook() == 'hi'
    assert time.time() - start >= 3

def test_cached_rich_args_rich_response(container):

    response = {}
    request = {}
    for i in range(400):
        response[random.randint(1, 1000)] = ['a', (2, 3), {'b': 4.3}]
        request[random.randint(1, 1000)] = ['b', [4, 6], {'c': 8.9}]

    def fake_some_method(*args, **kwargs):
        return response

    with patch('nameko.rpc.MethodProxy.__call__', fake_some_method):
        with entrypoint_hook(container, 'cached') as hook:
            assert hook(request) == response

    def broken_some_method(*args, **kwargs):
        raise Exception('hmm')

    with patch('nameko.rpc.MethodProxy.__call__', broken_some_method):
        with entrypoint_hook(container, 'cached') as hook:
            assert hook(request) == response

def test_cache_first(container):

    mock = Mock()
    with patch('nameko.rpc.MethodProxy.__call__', mock):
        with entrypoint_hook(container, 'cache_first') as hook:
            hook('ho')
    mock.assert_called_once_with('ho')

    mock.reset_mock()
    with patch('nameko.rpc.MethodProxy.__call__', mock):
        with entrypoint_hook(container, 'cache_first') as hook:
            hook('ho')
    mock.assert_not_called()

    cache_first_rpc = get_extension(container, CacheFirstRpcProxy)
    cache_first_rpc.cache = {}

    with patch('nameko.rpc.MethodProxy.__call__', mock):
        with entrypoint_hook(container, 'cache_first') as hook:
            hook('ho')
    mock.assert_called_once_with('ho')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
A few tools to cache interactions between your nameko services, increasing
resiliency and performance at the expense of consistency, when it makes sense.
"""


import json
from logging import getLogger

from eventlet.timeout import Timeout
from nameko.rpc import RpcProxy, ServiceProxy, MethodProxy


_log = getLogger(__name__)


class CachedRpcProxy(RpcProxy):
    """
    Stores responses from the original services and keeps them cached.

    If a cached version of this request exists, a response from the cache is
    sent instead of hangling forever or raising an exception.

    If a cached version doesn't exist, it will behave like a normal rpc,
    and wait indefinitey for a reply. All successful replies are cached.

    WARNING: Do NOT use this for setters, rpcs meant to modify state in the
    target service

    args:
      - cache: the cache to use. This should resemble a dict but can be more
        sophisticated, like the caches provided by the cachetools package.
      - failover_timeout: if a cached version of this query exists, how long
        in seconds should your original request wait until it deems the target
        service as unresponsive and moves on to use a cached response
    """

    def __init__(self, target_service, **options):
        self.target_service = target_service
        self.cache = options.pop('cache', dict())
        self.options = options

    def get_dependency(self, worker_ctx):
        return CachedServiceProxy(
            worker_ctx,
            self.target_service,
            self.rpc_reply_listener,
            cache=self.cache,
            **self.options
        )


class CacheFirstRpcProxy(RpcProxy):
    """
    Stores responses from the original services and keeps them cached.

    If further requests come in with the same arguments, a response from the
    cache is sent instead of hitting the destination service.

    WARNING: Do NOT use this for setters, rpcs meant to modify state in the
    target service

    args:
      - cache: the cache to use. This should resemble a dict but can be more
        sophisticated, like the caches provided by the cachetools package.
    """

    def __init__(self, target_service, **options):
        self.target_service = target_service
        self.cache = options.pop('cache', dict())
        options['use_cache_first'] = True
        self.options = options

    def get_dependency(self, worker_ctx):
        return CachedServiceProxy(
            worker_ctx,
            self.target_service,
            self.rpc_reply_listener,
            cache=self.cache,
            **self.options
        )


class CachedServiceProxy(ServiceProxy):

    def __init__(self, *args, **kwargs):
        cache = kwargs.pop('cache')
        super(CachedServiceProxy, self).__init__(*args, **kwargs)
        self.cache = cache.get(self.service_name, {})
        cache[self.service_name] = self.cache

    def __getattr__(self, name):
        return CachedMethodProxy(
            self.worker_ctx,
            self.service_name,
            name,
            self.reply_listener,
            cache=self.cache,
            **self.options
        )


class CachedMethodProxy(MethodProxy):

    def __init__(self, *args, **kwargs):
        self.cache = kwargs.pop('cache')
        self.failover_timeout = kwargs.pop('failover_timeout', 0)
        self.use_cache_first = kwargs.pop('use_cache_first', False)
        super(CachedMethodProxy, self).__init__(*args, **kwargs)

    def get_from_cache(self, key):
        if key not in self.cache:
            return False, None
        return True, self.cache[key][0]

    def __call__(self, *args, **kwargs):
        args_hash = json.dumps((args, kwargs), sort_keys=True)
        timeout = None
        incache, cached_response = self.get_from_cache(args_hash)
        if self.use_cache_first and incache:
            return cached_response
        if self.failover_timeout and incache:
            # we have a value, let's set a timeout and use our cache if service
            # isn't responsive
            timeout = Timeout(self.failover_timeout)
        try:
            reply = super(CachedMethodProxy, self).__call__(*args, **kwargs)
        except (Exception, Timeout) as e:
            error = 'timeout' if isinstance(e, Timeout) else 'error'
            _log.warn('%s when getting value for %s. using cache', error, self)
            if incache:
                _log.warn('response found in cache. using cache')
                return cached_response
            else:
                _log.error('request not in cache, re-raising %s', self.cache)
                raise e
        finally:
            if timeout:
                timeout.cancel()
        self.cache[args_hash] = (reply,)
        return reply



from .nameko_cachetools import CachedRpcProxy, CacheFirstRpcProxy

__author__ = 'Santiago Suarez Ordonez'
__email__ = 'santiycr@gmail.com'
__version__ = '1.0.0'
__all__ = ['CachedRpcProxy', 'CacheFirstRpcProxy']
